function simulator()

%%%%%%%%%%%%%
% Constants %
%%%%%%%%%%%%%

% Make these constants visible to external functions
global N dT PACKET_AGE_LIMIT POINT_R MAX_U CENTRAL_MOMENTS GOAL ...
    GAMMA A_COEFF B_COEFF U_GAINS

N = 9;         % number of bots
dT = .001;     % timestep length (position is updated each step)
SIDE = 800;    % side of bounding square for initial robot placement

% Rendering
DRAW_STEPS = 20;     % how many timesteps to wait between display updates
DRAW_CONNECTIONS = false;
DRAW_ESTIMATES = true;
DRAW_BOTS = true;
V_WHISKER_SCALE = 0; % if >0, controller outputs will be plotted as scaled arrows
RAINBOW_MODE = true; % give each bot its own color
RAINBOW_COLORS = [[1 0 0]; [1 .5 0]; [1 1 0]; [0 1 0]; [0 0 1];...
                [.3 0 .5]; [.9 .5 .9]]; %ROYGBIV
POINT_R = 25;        % distance from center to reference point
R = 50;              % radius of bots
WHEEL_OFFSET = 30;   % distance from wheels to center
WHEEL_R = 20;        % wheel radius

% Communication
COMM_DROPOFF = 700;    % distance where packet loss function hits 50%
PACKET_AGE_LIMIT = 200; % timesteps a recieved packet will be reused
                        %before dropping the bot

% Control flags
PERFECT_INFORMATION = false;  % feed perfect info. to controller instead of estimate
CENTRAL_MOMENTS = true;       % use central moments (not global moments)

% Control gains
A_COEFF = 140/N;     % consensus estimator
B_COEFF = 20/N;      % consensus estimator
GAMMA = 10;          % consensus input
U_GAINS = diag([10.0, 10.0, 0.0001, 0.0002, 0.0001]);  % controller

% Update cycle
UPDATE_STEPS = 10; % timesteps an agent will wait between update cycles
UPDATE_RAND = 4;   % randomness in update wait times
                   %(0 means all agents update synchronously)

% Speed constraints
MAX_U = 1000;                 % speed limit (physical constraint of the e-pucks)
COLLISION_RAD = 130; % see collision avoidance

% Goal
GOAL = [100 300 160000 40000 40000]';
if ~CENTRAL_MOMENTS
    GOAL = [GOAL(1)
            GOAL(2)
            GOAL(3) + GOAL(1)^2
            GOAL(4) + GOAL(1)*GOAL(2);
            GOAL(5) + GOAL(2)^2];
end

% Movie recording
MOVIE_FRAME_COUNT = 0;    % number of frames to be captured @30fps


%%%%%%%%%%%%%
% Main loop %
%%%%%%%%%%%%%

bots = distributeBots(N, rand()*(2000-SIDE)-(1000-SIDE/2),...
    rand()*(2000-SIDE)-(1000-SIDE/2));

mainFigure = figure( ...
   'Position',[50, 50, 600, 600], ...
   'color', [1 1 1], ...
   'Menubar', 'none', ...
   'Name','Swarm Formation', ...
   'DoubleBuffer', 'on', ...
   'WindowButtonDownFcn', @mouseListener, ...
   'DeleteFcn', @windowKiller);

if MOVIE_FRAME_COUNT > 0
    fprintf(2, 'RECORDING MOVIE.\n');
    aviobj = avifile('movie.avi', 'compression', 'none', ...
        'quality', 100, 'fps', 30); % ~650 KB/frame
end

stepCounter = 0;
keepLooping = true;
while keepLooping
    
    %Draw
    if mod(stepCounter, DRAW_STEPS) == 0
        drawField(mainFigure, bots);
        drawnow;
        
        if MOVIE_FRAME_COUNT > 0
            aviobj = addframe(aviobj, getframe(gca));
            if stepCounter >= (MOVIE_FRAME_COUNT-1)*DRAW_STEPS
                aviobj = close(aviobj);
                close(gcf);
                display(aviobj);
                fprintf(2, 'MOVIE SUCCESSFULLY RECORDED.\n');
                break;
            end
        end
    end
    
    %Run bots' update cycles if it's time
    for i=1:N
        if bots(i).lastUpdate+bots(i).waitTime <= stepCounter
            bots = transmitPacket(bots, i);
            bots = updateBotComputations(bots, i);
            bots(i).lastUpdate = stepCounter;
            bots(i).waitTime = generateWaitTime();
        end
    end
    
    %Update plant and timers
    for i=1:N
        bots(i).state = plant(bots(i));
        
        for g=1:N
            bots(i).packets(g).age = bots(i).packets(g).age+1;
        end
        bots(i).packets(i).age = 0;
    end
    stepCounter = stepCounter+1;
end

% Redistribute bots with mouse click
    function mouseListener(src, eventdata)
        p = get(gca,'currentpoint');
        bots = distributeBots(N, p(1), p(3));
        stepCounter = 0;
    end
% Close window
    function windowKiller(src, eventdata)
        keepLooping = false;
    end

%%%%%%%%%%%%%%%%%%%%%%
% Bot initialization %
%%%%%%%%%%%%%%%%%%%%%%
    function bots = distributeBots(N, xStart, yStart)
        bots(N,1) = struct();
        for j=1:N
            % place agent
            badLocation = true;
            for k=1:(N*100)
                p = [xStart + rand()*SIDE-SIDE/2; yStart + rand()*SIDE-SIDE/2];
                minDist = Inf;
                for m=1:(j-1)
                    minDist = min(minDist, sqrt(sum((p - bots(m).state.p).^2)));
                end
                if minDist > COLLISION_RAD
                    badLocation = false;
                    break;
                end
            end
            if badLocation
                error('Could not fit robots into starting grid');
            end
            
            % set up misc. fields
            bots(j).state.p = p;
            bots(j).state.angle = rand()*2*pi;
            
            packet = struct('age', PACKET_AGE_LIMIT+1);
            packet.v = zeros(5,1);
            packet.w = zeros(5,1);
            for k=1:N
                bots(j).packets(k) = packet;
            end
            bots(j).packets(j).v = calcEllipse(p);
            bots(j).packets(j).w = zeros(5,1);
            bots(j).u = [0; 0];
            bots(j).lastUpdate = 0;
            bots(j).waitTime = generateWaitTime();
        end
    end

%%%%%%%%%%%%%%%%
% Update Cycle %
%%%%%%%%%%%%%%%%

    function bots = updateBotComputations(bots, n)
        dt = dT*bots(n).waitTime;
        
        sensory = sensors(bots(n).state, 0);
        
        packet = consensusEstimator(bots(n).packets, n);
        packet = consensusInput(bots(n).packets(n), packet, sensory);
        
        packet.v = bots(n).packets(n).v + packet.v*dt;
        packet.w = bots(n).packets(n).w + packet.w*dt;
        bots(n).packets(n) = packet;
        
        if PERFECT_INFORMATION
            s = [bots.state];
            packet.v = calcEllipse([s.p]);
        end
        controls = controller(sensory, packet);
        
        bots(n).u = avoidCollisions([bots.state], n, controls);
    end

% Generate a waiting period between until next update (in timesteps)
    function steps = generateWaitTime()
        steps = UPDATE_STEPS + (rand()*2-1)*UPDATE_RAND;
    end

% Transmit bot n's packet to any neighbors for whom packetLost returns false
    function bots = transmitPacket(bots, n)
        for j=1:N
            if ~packetLost(norm(bots(n).state.p - bots(j).state.p))
                bots(j).packets(n) = bots(n).packets(n);
            end
        end
    end

% cumulative Rice function for packet loss probabilities
    function p = cumuRice(x)
        x = x*2.2457;
        alpha = 2/1;
        beta = x/1;
        m = 0:30; %increase upper bound for greater precision
        bessels = besseli(m, alpha*beta);
        p = 1-exp(-(alpha^2+beta^2)/2)*sum((alpha/beta).^m .* bessels);
    end
% boolean -- get random sample from Rice function at distance x
    function p = packetLost(x)
        if x == 0 || COMM_DROPOFF == Inf
            p = 0;
        else
            p = rand() < cumuRice(x/COMM_DROPOFF);
        end
    end

%%%%%%%%%%%%%%%%%
% Misc. Helpers %
%%%%%%%%%%%%%%%%%


% Collision avoidance using distance-based velocity scaling (scale by alpha)
%along the vector between an agent and the agent nearest to it
%        1 |          _______
%  alpha   |        /
%          |      /
%        0 |___ /
%          0   R2    R1    ++++
%            distance
    function u = avoidCollisions(states, n, u)
        R1 = COLLISION_RAD*3;
        R2 = COLLISION_RAD;
        dist = Inf;
        for j = 1:N
            d = norm(states(n).p-states(j).p);
            if j ~= n && d < dist
                k = j;
                dist = d;
            end
        end
        pDiff = states(n).p - states(k).p;
        
        if dist > R2
            alpha = (dist-R2)/(R1-R2);
        else
            alpha = 0;
        end
        if dot(u, pDiff) < 0
            v2 = project(u, [pDiff(2); -pDiff(1)]);
            u = alpha*u + (1-alpha)*v2;
        end
    end

% Vector projection of b onto a
    function v = project(b, a)
        v = dot(a, b)/sum(a.^2)*a;
    end

% Get the center point of the robot (based on angle and register point)
    function p = calcCenterPoint(state)
        p = state.p - [cos(state.angle); sin(state.angle)]*POINT_R;
    end

%%%%%%%%%%%%%
% Rendering %
%%%%%%%%%%%%%
    function drawField(fig, bots)
        clf(fig);
        figure(fig);
        axes( ...
          'xlim', [-1200 1200], 'ylim', [-1200 1200], ...
          'XTick', -1000:200:1000, 'YTick', -1000:200:1000, ...
          'dataaspectratio',[1 1 1]);
        grid on;
        hold on;

        if DRAW_CONNECTIONS
            for j=1:N
                for k=1:N
                    age = min(bots(j).packets(k).age, bots(k).packets(j).age);
                    if age < PACKET_AGE_LIMIT
                        if age <= max(bots(j).waitTime, bots(k).waitTime)
                            color = [0 0 0];
                        else
                            color = [1 .5 .5];
                        end
                        line('XData', [bots(j).state.p(1), bots(k).state.p(1)], ...
                            'YData', [bots(j).state.p(2), bots(k).state.p(2)], ...
                            'Color', color);
                    end
                end
            end
        end

        for j=1:N
            if RAINBOW_MODE
                %add a new, random color if enough unique ones don't exist
                if j > length(RAINBOW_COLORS)
                    RAINBOW_COLORS(j,:) = [rand(), rand(), rand()];
                end
                color = RAINBOW_COLORS(j,:);
            else
                color = [1 0 0];
            end
            line(bots(j).state.p(1), bots(j).state.p(2), 'Marker', 'o', ...
                'MarkerFaceColor',color, 'MarkerEdgeColor',[0 0 0],'MarkerSize',5);
            if DRAW_ESTIMATES
                drawEllipse_global(bots(j).packets(j).v, color);
            end
            if DRAW_BOTS
                p = calcCenterPoint(bots(j).state);
                angle = bots(j).state.angle;
                x = p(1) + R*cos(linspace(0, 2*pi, 16));
                y = p(2) + R*sin(linspace(0, 2*pi, 16));
                line(x, y, 'Color', [.5 .5 .5]);
                wheelOff = [sin(angle); -cos(angle)]*WHEEL_OFFSET;
                wheelR = [cos(angle); sin(angle)]*WHEEL_R;
                line(p(1)+wheelOff(1) + [wheelR(1); -wheelR(1)], ...
                    p(2)+wheelOff(2) + [wheelR(2); -wheelR(2)], ...
                    'Color', [.5 .5 .5]);
                line(p(1)-wheelOff(1) + [wheelR(1); -wheelR(1)], ...
                    p(2)-wheelOff(2) + [wheelR(2); -wheelR(2)], ...
                    'Color', [.5 .5 .5]);
            end
            if V_WHISKER_SCALE > 0
                p0 = bots(j).state.p;
                off = bots(j).u*V_WHISKER_SCALE;
                quiver(p0(1), p0(2), off(1), off(2));
            end
        end

        s = [bots.state];
        drawEllipse_global(calcEllipse([s.p]), [0 0 0]);
        if CENTRAL_MOMENTS
            drawEllipse_central(GOAL, [.2 .5 .2]);
        else
            drawEllipse_global(GOAL, [.2 .5 .2]);
        end

        function drawEllipse_global(e, color)
            drawEllipse_central( [  e(1)
                                    e(2)
                                    e(3) - e(1)^2
                                    e(4) - e(1)*e(2)
                                    e(5) - e(2)^2], ...
                                 color);
        end
        function drawEllipse_central(e, color)
            K = 50; %vertex count
            
            [u,s,v] = svd([e(3) -e(4); -e(4) e(5)]);

            H = patch('edgecolor',color, ...
                'facecolor','none','LineWidth',2);

            % make the circle to an ellipse
            xpts = sqrt(4*s(1,1))*cos(linspace(0, 2*pi, K));
            ypts = sqrt(4*s(2,2))*sin(linspace(0, 2*pi, K));
            set(H, 'XData', xpts, 'YData', ypts);
            rotate(H, [0 0 1], atan2(u(1,2),u(2,2))*180/pi, [0 0 0]);
            set(H, 'XData', get(H,'XData')+e(1), 'YData', get(H,'YData')+e(2));

            % draw the little center dot
            line(e(1), e(2), 'marker', 'o', ...
                'markerfacecolor','none', 'markeredgecolor',color,'markersize',10);
        end

    end

end